use std::marker::PhantomData;
use crate::{
    prelude::*,
    runtime,
    types,
    Any,
};

pub struct Array<const N :usize, T: SzunType> {
    address:Ref,
    _data:PhantomData<T>,
}
impl<const N :usize, T: SzunType> Array<N,T> {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())}, _data:PhantomData }
    }

    pub fn set<D: SzunType>(&mut self, index:usize, data:&D)
    {
        if self.compatible(data) {
            unsafe {runtime::array_set(self.address.clone(), index, data.__to_ref())};
        }
    }

    pub fn get(&self, index:usize) -> T
    {
        let refr = unsafe {runtime::array_at(self.address.clone(), index)};
        if refr.is_some() {
            T::__from_ref(refr)
        } else { panic!("array index out of bounds"); }
    }

    pub fn len(&self) -> usize
    {
        N
    }
}
impl<const N :usize, T: SzunType> SzunType for Array<N,T> {
    fn type_id() -> usize
    {
        let inner = unsafe {runtime::type_outer(T::type_id(), N)};
        unsafe {runtime::type_outer(inner, types::TYPE_ARRAY)}
    }

    fn compatible<R: SzunType>(&self, _:&R) -> bool
    {
        R::type_id() == T::type_id() || unsafe {runtime::type_key(T::type_id())} == types::TYPE_VAR
    }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr, _data:PhantomData } }
}
impl<const N :usize, T: SzunType> Clone for Array<N,T> {
    fn clone(&self) -> Self
    {
        Self::__from_ref(unsafe {runtime::clone(self.address.clone())})
    }
}
impl<const N :usize, T: SzunType> TryFrom<Any> for Array<N,T> {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<const N :usize, T: SzunType> Into<Any> for Array<N,T> { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl<const N :usize, T: SzunType> Drop for Array<N,T> {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}
