use crate::{
    prelude::*,
    runtime,
    types,
};

pub struct Boolean {
    address:Ref,
}
impl Boolean {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())} }
    }

    pub fn with(mut self, data:bool) -> Self
    {
        self.set(data);
        self
    }

    pub fn set(&mut self, data:bool)
    {
        unsafe {runtime::bool_set(self.address.clone(), data)}
    }
    
    pub fn get(&self) -> bool
    {
        unsafe {runtime::bool_get(self.address.clone())}
    }
}
impl SzunType for Boolean {
    fn type_id() -> usize { unsafe {runtime::type_outer(0, types::TYPE_BOOLEAN)} }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }
    
    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}
impl Clone for Boolean {
    fn clone(&self) -> Self
    {
        Self::__from_ref(unsafe {runtime::clone(self.address.clone())})
    }
}
impl TryFrom<Any> for Boolean {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Any> for Boolean { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl Drop for Boolean {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}

pub fn boolean(data:bool) -> Boolean
{
    Boolean::from(data)
}

impl std::ops::Not for Boolean {
    type Output = Self;
    fn not(self) -> Self::Output
    {
        Self::from(!self.get())
    }
}

impl From<bool> for Boolean { fn from(value: bool) -> Self { Self::new().with(value) } }
impl Into<bool> for Boolean { fn into(self) -> bool { self.get() } }
