use crate::{
    prelude::*,
    runtime,
    types,
};

pub struct Integer {
    address:Ref,
}
impl Integer {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())} }
    }

    pub fn with<T: num::Num + num::NumCast>(mut self, data:T) -> Self
    {
        self.set(data);
        self
    }

    pub fn set<T: num::Num + num::NumCast>(&mut self, data:T)
    {
        match num::NumCast::from(data) {
            Some(data) => unsafe {runtime::integer_set(self.address.clone(), data)}
            None => { }
        }
    }
    
    pub fn get<T: num::Num + num::NumCast>(&self) -> T
    {
        num::NumCast::from(unsafe {runtime::integer_get(self.address.clone())}).unwrap()
    }
}
impl SzunType for Integer {
    fn type_id() -> usize { unsafe {runtime::type_outer(0, types::TYPE_INTEGER)} }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}
impl Clone for Integer {
    fn clone(&self) -> Self
    {
        Self::__from_ref(unsafe {runtime::clone(self.address.clone())})
    }
}
impl TryFrom<Any> for Integer {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Any> for Integer { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl Drop for Integer {
    fn drop(&mut self)
    {
        unsafe {runtime::release(self.address.clone())};
    }
}

impl std::ops::Neg for Integer {
    type Output = Self;
    fn neg(self) -> Self::Output
    {
        Self::new().with(-self.get::<i64>())
    }
}
crate::util::macros::impl_ops_numeric_all!(Integer, i64);

pub fn integer<T: num::Num + num::NumCast>(data:T) -> Integer
{
    Integer::new().with(data)
}
