use crate::{
    prelude::*, types,
    runtime,
};

pub struct Natural {
    address:Ref,
}
impl Natural {
    pub fn new() -> Self
    {
        Self { address:unsafe {runtime::acquire(Self::type_id())} }
    }

    pub fn with<T: num::Num + num::NumCast>(mut self, data:T) -> Self
    {
        self.set(data);
        self
    }

    pub fn set<T: num::Num + num::NumCast>(&mut self, data:T)
    {
        match num::NumCast::from(data) {
            Some(data) => unsafe {runtime::natural_set(self.address.clone(), data)}
            None => { }
        }
    }
    
    pub fn get<T: num::Num + num::NumCast>(&self) -> T {
        num::NumCast::from(unsafe {runtime::natural_get(self.address.clone())}).unwrap()
    }
}
impl SzunType for Natural {
    fn type_id() -> usize { unsafe {runtime::type_outer(0, types::TYPE_NATURAL)} }

    fn compatible<T: SzunType>(&self, _:&T) -> bool { false }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr } }
}
impl Clone for Natural {
    fn clone(&self) -> Self {
        Self::__from_ref(unsafe {runtime::clone(self.address.clone())})
    }
}
impl TryFrom<Any> for Natural {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl Into<Any> for Natural { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl Drop for Natural {
    fn drop(&mut self) {
        unsafe {runtime::release(self.address.clone())};
    }
}

crate::util::macros::impl_ops_numeric_all!(Natural, u64);

pub fn natural<T: num::Num + num::NumCast>(data:T) -> Natural
{
    Natural::new().with(data)
}
