use std::marker::PhantomData;
use crate::{
    prelude::*,
    runtime,
    types,
};

pub struct Table<T: SzunType> {
    address:Ref,
    _data:PhantomData<T>,
}
impl<T: SzunType> Table<T> {
    pub fn new() -> Self
    {
        Self {
            address:unsafe {runtime::acquire(Self::type_id())},
            _data:PhantomData
        }
    }

    //pub fn with_size(mut self, length:usize) -> Self { }

    //pub fn set<R: SzunType>(&mut self, index:usize, data:&R) { }
    //pub fn unset(&mut self, index:usize) { }
    //pub fn get(&self, index:usize) -> T { }
    //pub fn try_get(&self, index:usize) -> Result<T,()> { }
    //pub fn resize(&mut self, length:usize) { }
    //pub fn clear(&mut self) { }
    //pub fn len(&self) -> usize { 0 }
    //pub fn next(&self) -> usize { 0 }
}
impl<T: SzunType> SzunType for Table<T> {
    fn type_id() -> usize {
        unsafe {runtime::type_outer(T::type_id(), types::TYPE_TABLE)}
    }

    fn compatible<R: SzunType>(&self, _:&R) -> bool {
        R::type_id() == T::type_id() || unsafe {runtime::type_key(T::type_id())} == types::TYPE_VAR
    }

    fn __to_ref(&self) -> Ref { self.address.clone() }
    fn __from_ref(refr:Ref) -> Self { Self { address:refr, _data:PhantomData } }
}
impl<T: SzunType> Clone for Table<T> {
    fn clone(&self) -> Self { Self::__from_ref(unsafe {runtime::clone(self.address.clone())}) }
}
impl<T: SzunType> TryFrom<Any> for Table<T> {
    type Error = SzunError;
    fn try_from(obj: Any) -> Result<Self, Self::Error>
    {
        if obj.type_id() == Self::type_id() {
            Ok(Self::__from_ref(obj.__to_ref()))
        } else { Err(SzunError::MismatchedType) }
    }
}
impl<T: SzunType> Into<Any> for Table<T> { fn into(self) -> Any { Any::__from_ref(self.address.clone()) } }
impl<T: SzunType> Drop for Table<T> {
    fn drop(&mut self) { unsafe {runtime::release(self.address.clone())}; }
}
