use crate::{
    prelude::*,
    runtime::type_key,
};

#[repr(C)]
#[derive(Clone, Copy)]
pub struct Ref {
    pub(crate) class:usize,
    pub(crate) address:usize,
}
impl Ref {
    pub fn null() -> Self { Self { class:0, address:0 } }

    pub fn is_some(&self) -> bool { self.address != 0 }
    pub fn is_none(&self) -> bool { self.address == 0 }

    pub fn type_id(&self) -> usize { self.class }

    pub fn is<T: SzunType>(&self) -> bool { unsafe {type_key(T::type_id()) == type_key(self.class)} }

    pub(crate) fn clone(&self) -> Self
    {
        Ref { class:self.class, address:self.address }
    }

    pub fn try_into<T: SzunType>(self) -> Result<T, SzunError>
    {
        if self.class == T::type_id() {
            Ok(T::__from_ref(self))
        } else {
            Err(SzunError::MismatchedType)
        }
    }
}

