#include "lib.h"

Reference list_cell(Reference addr, size_t index)
// Returns a reference to the list cell at the specified index.
//
{
    Reference result {0};
    Type::List& list = *reinterpret_cast<Type::List*>(addr.address);
    size_t inner = type_inner(addr.type);
    size_t offset = type_size(inner);

    // validate for overflow
    if(list.data.data != nullptr && offset > 0 && index < list.data.capacity) {
        result.type = inner;
        result.address = rawlist_cell(list.data, offset, index);
    }

    return result;
}


extern "C" size_t list_capacity(Reference addr)
// Returns the capacity of the list.
//
{
    return (*reinterpret_cast<Type::List*>(addr.address)).data.capacity;
}

extern "C" size_t list_length(Reference addr)
// Returns the number of elements in the list.
//
{
    return (*reinterpret_cast<Type::List*>(addr.address)).data.length;
}

extern "C" Reference list_at(Reference addr, size_t index)
// Returns a reference to the cell at the specified index.
//
{
    Reference result {0};
    Reference cell = list_cell(addr, index);
    
    if(cell.address != nullptr) {
        //if(type_key(cell.type) == Type::TYPE_VARYING) {
        //    result = *reinterpret_cast<Reference*>(cell.address);
        //} else {
            result = cell;
        //}
    }
    return refr_make_temporary(result);
}

extern "C" Reference list_front(Reference addr)
// Returns a reference to the first element in the list.
//
{
    return list_at(addr, 0);
}

extern "C" Reference list_back(Reference addr)
// Returns a reference to the last element in the list.
//
{
    Reference result {0};
    Type::List& list = *reinterpret_cast<Type::List*>(addr.address);
    if(list.data.length > 0) {
        result = list_at(addr, list.data.length - 1);
    }
    return result;
}

extern "C" void list_clear(Reference addr)
// Drops and removes all elements from the list.
//
{
    Type::List& list = *reinterpret_cast<Type::List*>(addr.address);
    for(size_t i = 0; i < list.data.length; ++i) {
        drop(list_cell(addr, i));
    }
    rawlist_clear(list.data);
}

extern "C" void list_insert(Reference addr, size_t index, Reference source)
// Inserts an element at the specified index, shifting the trailing items.
//
{
    auto& list = (*reinterpret_cast<Type::List*>(addr.address));
    size_t inner = type_inner(addr.type);
    size_t offset = type_size(inner);

    if(index > list.data.length) { index = list.data.length; }

    Reference cell {0};
    cell.type = inner;
    cell.address = rawlist_insert(list.data, offset, index);

    if(type_key(inner) == Type::TYPE_VARYING) {
        var_set(cell, source);
    }
    else {
        copy(cell, source);
    }
}

extern "C" void list_push(Reference addr, Reference source)
// Adds an element to the end of the list.
//
{
    auto& list = *reinterpret_cast<Type::List*>(addr.address);
    list_insert(addr, list.data.length, source);
}

extern "C" void list_pop(Reference addr)
// Removes an element from the end of the list.
//
{
    auto& list = *reinterpret_cast<Type::List*>(addr.address);
    list_remove(addr, list.data.length - 1);
}

extern "C" void list_set(Reference addr, size_t index, Reference source)
{
    auto& list = *reinterpret_cast<Type::List*>(addr.address);
    if(index < list.data.length) {
        size_t inner = type_inner(addr.type);

        if(type_key(inner) == Type::TYPE_VARYING) {
            var_set(list_cell(addr, index), source);
        }
        else {
            copy(list_cell(addr, index), source);
        }
    }
}

extern "C" void list_remove(Reference addr, size_t index)
{
    auto& list = *reinterpret_cast<Type::List*>(addr.address);
    size_t inner = type_inner(addr.type);
    size_t offset = type_size(inner);
    if(index < list.data.length) {
        drop(list_at(addr, index));
        rawlist_remove(list.data, offset, index);
    }
}

extern "C" void list_reserve(Reference addr, size_t capacity)
{
    auto& list = *reinterpret_cast<Type::List*>(addr.address);
    size_t inner = type_inner(addr.type);
    size_t offset = type_size(inner);

    rawlist_reserve(list.data, offset, capacity);
}
