#include "lib.h"

extern "C" size_t record_length(Reference addr)
{
    auto binding = DB_SCHEMA.get(type_innerkey(addr.type));
    if(binding != nullptr) {
        return binding->data.size();
    }
    return 0;
}

extern "C" size_t record_type(Reference addr, size_t index)
{
    auto binding = DB_SCHEMA.get(type_innerkey(addr.type));
    if(binding != nullptr && index < binding->data.size()) {
        return binding->data[index].type;
    }
    return 0;
}

Reference record_cell(Reference addr, size_t index)
{
    Reference result {0};
    auto binding = DB_SCHEMA.get(type_innerkey(addr.type));
    if(binding != nullptr) {
        if(index < binding->data.size()) {
            result.type = binding->data[index].type;
            result.address = addr.address + binding->data[index].offset;
        }
    }
    return result;
}

extern "C" Reference record_at(Reference addr, size_t index)
{
    Reference result = record_cell(addr, index);
    if(type_key(result.type) == Type::TYPE_VARYING) {
        result = var_get(result);
    }
    return result;
}

extern "C" void record_update(Reference addr, size_t index, Reference source)
{
    Reference destination {0};
    size_t schid = type_innerkey(addr.type);
    auto binding = DB_SCHEMA.get(schid);
    if(binding != nullptr) {
        if(index < binding->data.size()) {
            destination.type = binding->data[index].type;
            destination.address = addr.address + binding->data[index].offset;
            copy(destination, source);
        }
    }
}

extern "C" size_t record_keyof(Reference addr, size_t index)
{
    Reference destination {0};
    auto binding = DB_SCHEMA.get(type_innerkey(addr.type));
    if(binding != nullptr) {
        if(index < binding->data.size()) {
            return binding->data[index].key;
        }
    }
    return 0;
}

extern "C" size_t record_indexof(Reference addr, size_t key)
{
    Reference destination {0};
    auto binding = DB_SCHEMA.get(type_innerkey(addr.type));
    if(binding != nullptr) {
        auto result = binding->map.get(key);
        if(result != nullptr) {
            return *result;
        }
        return binding->data.size();
    }
    return 0;
}
