#include "lib.h"

extern "C" size_t schema_length(Reference addr)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    return object.data.length;
}

extern "C" size_t schema_insert(Reference addr, size_t index, size_t type_id)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    if(index >= object.data.length) {
        index = object.data.length;
    }
    void* cell = rawlist_insert(object.data, sizeof(size_t), index);
    *reinterpret_cast<size_t*>(cell) = type_id;
    return index;
}

extern "C" void schema_update(Reference addr, size_t index, size_t type_id)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    if(index < object.data.length) {
        auto cell = reinterpret_cast<size_t*>(rawlist_cell(object.data, sizeof(size_t), index));
        if(cell != nullptr) {
            *cell = type_id;
        }
    }
}

extern "C" size_t schema_get(Reference addr, size_t index)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    if(index < object.data.length) {
        auto cell = reinterpret_cast<size_t*>(rawlist_cell(object.data, sizeof(size_t), index));
        if(cell != nullptr) {
            return *cell;
        }
    }
    return 0;
}

extern "C" void schema_data_set(Reference addr, size_t index)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    if(index < object.data.length) {
        auto cell = reinterpret_cast<size_t*>(rawlist_cell(object.map, sizeof(size_t), index));
        if(cell != nullptr) {
            
        }
    }
}

extern "C" void schema_data_get(Reference addr, size_t index)
{
    
}

extern "C" void schema_remove(Reference addr, size_t index)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    if(index < object.data.length) {
        rawlist_remove(object.data, sizeof(size_t), index);

        // remove mapping
        for(size_t i = 0; i < object.map.length; i++) {
            auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.data, sizeof(Type::Schema::Mapping), i));
            if(cell != nullptr) {
                if(cell->index == index) {
                    rawlist_remove(object.map, sizeof(Type::Schema::Mapping), i);
                    return;
                }
            }
        }
    }
}

extern "C" void schema_clear(Reference addr)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    rawlist_clear(object.data);
    rawlist_clear(object.map);
}

extern "C" void schema_map(Reference addr, size_t key, size_t index)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    if(index < object.data.length) {
        size_t find_index = 0;
        for(; find_index < object.map.length; find_index++) {
            auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.map, sizeof(Type::Schema::Mapping), find_index));
            if(cell != nullptr) {
                if(cell->key == key) { break; }
            }
        }

        // if key is not found, add new mapping
        if(find_index == object.map.length) {
            auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_insert(object.map, sizeof(Type::Schema::Mapping), object.map.length));
            cell->key = key;
            cell->index = index;
        }
        // otherwise, update existing key
        else {
            auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.map, sizeof(Type::Schema::Mapping), find_index));
            cell->index = index;
        }
    }
}

extern "C" void schema_unmap(Reference addr, size_t key)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    for(size_t i; i < object.map.length; ++i) {
        auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.map, sizeof(Type::Schema::Mapping), i));
        if(cell != nullptr) {
            if(cell->key == key) {
                rawlist_remove(object.map, sizeof(Type::Schema::Mapping), i);
                return;
            }
        }
    }
}

extern "C" size_t schema_indexof(Reference addr, size_t key)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    for(size_t i = 0; i < object.map.length; i++) {
        auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.map, sizeof(Type::Schema::Mapping), i));
        if(cell != nullptr) {
            if(cell->key == key) {
                return cell->index;
            }
        }
    }
    return object.data.length;
}

extern "C" size_t schema_keyof(Reference addr, size_t index)
{
    auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));
    for(size_t i = 0; i < object.map.length; i++) {
        auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.map, sizeof(Type::Schema::Mapping), i));
        if(cell != nullptr) {
            if(cell->index == index) {
                return cell->key;
            }
        }
    }
    return object.data.length;
}

extern "C" size_t schema_bind(Reference addr, size_t id)
{
    if(id > 0) {
        Type::SchemaBinding binding {0};
        auto& object = (*reinterpret_cast<Type::Schema*>(addr.address));

        // prepare binding
        binding.binding = id;
        for(size_t i = 0; i < object.data.length; ++i) {
            Type::SchemaBinding::Row row {0};

            size_t type_id = *reinterpret_cast<size_t*>(rawlist_cell(object.data, sizeof(size_t), i));
            size_t size = type_size(type_id);

            // fail if size is not valid
            if(size == 0) { return 0; }
            
            size_t alignment = type_alignment(type_id);
            binding.alignment = std::max(alignment, binding.alignment);
            size_t position = ((binding.size + (alignment - 1)) & ~(alignment - 1));
            binding.size = size + position;

            row.type = type_id;
            row.offset = position;
            binding.data.push_back(row);
        }
        binding.size = ((binding.size + (binding.alignment - 1)) & ~(binding.alignment - 1));
        binding.references = 0;

        for(size_t i = 0; i < object.map.length; ++i) {
            auto cell = reinterpret_cast<Type::Schema::Mapping*>(rawlist_cell(object.map, sizeof(Type::Schema::Mapping), i));
            binding.map.set(cell->key, cell->index);
            binding.data[cell->index].key = cell->key;
        }

        /* printf("[Binding]\n");
        printf("  Id:    %zu\n", binding.binding);
        printf("  Size:  %zu\n", binding.size);
        printf("  Align: %zu\n", binding.alignment);
        printf("  Data:\n");
        for(size_t i = 0; i < binding.data.size(); ++i) {
            printf("   - %zu:%#x +%zu\n",
                binding.data[i].type,
                type_key(binding.data[i].type),
                binding.data[i].offset
            );
        }
        printf("  Map:\n");
        for(size_t key : binding.map.indices()) {
            printf("   - %zu -> %zu\n",
                key,
                *binding.map.get(key)
            );
        } */

        // add binding to pool
        DB_SCHEMA.set(id, binding);
        return id;
    }
    return 0;
}

extern "C" bool schema_has(size_t id)
{
    return DB_SCHEMA.has(id);
}
