pub mod util;

use crate::prelude::Ref;

#[repr(C)]
#[derive(Clone, Copy)]
pub struct Str {
    pub bytes:*mut u8,
    pub length:usize,
}

#[repr(C)]
#[derive(Clone, Copy)]
pub struct SparseHeader {
    pub start:usize,
    pub length:usize,
    pub index:usize,
}

extern "C" {
    pub fn test();

    pub fn acquire(type_id:usize) -> Ref;
    pub fn release(addr:Ref);

    pub fn clone(src:Ref) -> Ref;
    pub fn const_clone(src:Ref) -> Ref;
    pub fn copy(dst:Ref, src:Ref) -> bool;
    pub fn transfer(dst:Ref, src:Ref) -> bool;

    pub fn type_outer(type_id:usize, key:usize) -> usize;
    pub fn type_inner(type_id:usize) -> usize;
    pub fn type_key(type_id:usize) -> usize;
    pub fn type_innerkey(type_id:usize) -> usize;
    pub fn kind_hasinner(kind:usize) -> usize;
    pub fn type_hasinner(type_id:usize) -> usize;
    pub fn type_sanitize(type_id:usize) -> usize;

    //pub fn type_size(type_id:usize) -> usize;

    pub fn name_indexof(data:*const u8, length:usize) -> usize;
    pub fn name_keyof(index:usize) -> Str;
    pub fn name_release(data:Str);

    pub fn bool_set(addr:Ref, data:bool);
    pub fn bool_get(addr:Ref) -> bool;
    
    pub fn natural_set(addr:Ref, data:u64);
    pub fn natural_get(addr:Ref) -> u64;

    pub fn integer_set(addr:Ref, data:i64);
    pub fn integer_get(addr:Ref) -> i64;

    //pub fn decimal_set(addr:Ref, data:i64);
    //pub fn decimal_get(addr:Ref) -> i64;

    //pub fn significant_set(addr:Ref, data:f64);
    //pub fn significant_get(addr:Ref) -> f64;
    
    pub fn block_get(addr:Ref, index:usize) -> u8;
    pub fn block_set(addr:Ref, index:usize, data:u8);
    pub fn block_get_bits(addr:Ref, index:usize, length:usize) -> usize;
    pub fn block_set_bits(addr:Ref, index:usize, length:usize, data:usize);

    pub fn sequence_length(addr:Ref) -> usize;
    pub fn sequence_get(addr:Ref, index:usize) -> u8;
    pub fn sequence_clear(addr:Ref);
    pub fn sequence_set(addr:Ref, index:usize, value:u8);
    pub fn sequence_resize(addr:Ref, capacity:usize);

    pub fn var_get(addr:Ref) -> Ref;
    pub fn var_set(addr:Ref, source:Ref);
    pub fn var_clear(addr:Ref);

    pub fn array_at(addr:Ref, index:usize) -> Ref;
    pub fn array_set(addr:Ref, index:usize, source:Ref);

    pub fn list_capacity(addr:Ref) -> usize;
    pub fn list_length(addr:Ref) -> usize;
    pub fn list_front(addr:Ref) -> Ref;
    pub fn list_back(addr:Ref) -> Ref;
    pub fn list_at(addr:Ref, index:usize) -> Ref;
    pub fn list_clear(addr:Ref);
    pub fn list_insert(addr:Ref, index:usize, src:Ref);
    pub fn list_push(addr:Ref, src:Ref);
    pub fn list_pop(addr:Ref);
    pub fn list_set(addr:Ref, index:usize, src:Ref);
    pub fn list_remove(addr:Ref, index:usize);
    pub fn list_reserve(addr:Ref, capacity:usize);

    pub fn sparse_length(addr:Ref) -> usize;
    pub fn sparse_clear(addr:Ref);
    pub fn sparse_first(addr:Ref) -> usize;
    pub fn sparse_next(addr:Ref) -> usize;
    pub fn sparse_last(addr:Ref) -> usize;
    pub fn sparse_get(addr:Ref, index:usize) -> Ref;
    pub fn sparse_at(addr:Ref, index:usize) -> Ref;
    pub fn sparse_set(addr:Ref, index:usize, source:Ref);
    pub fn sparse_unset(addr:Ref, index:usize);

    pub fn table_length(addr:Ref) -> usize;
    pub fn table_resize(addr:Ref, size:usize);
    pub fn table_clear(addr:Ref);
    pub fn table_set(addr:Ref, index:usize, data:Ref);
    pub fn table_unset(addr:Ref, index:usize);
    pub fn table_get(addr:Ref, index:usize) -> Ref;
    pub fn table_next(addr:Ref) -> usize;

    //pub fn record_length(addr:Ref) -> usize;
    //pub fn record_type(addr:Ref, index:usize) -> usize;
    //pub fn record_at(addr:Ref, index:usize) -> Ref;
    //pub fn record_update(addr:Ref, index:usize, source:Ref);
    //pub fn record_keyof(addr:Ref, index:usize) -> usize;
    //pub fn record_indexof(addr:Ref, key:usize) -> usize;

    pub fn schema_length(addr:Ref) -> usize;
    pub fn schema_insert(addr:Ref, index:usize, type_id:usize) -> usize;
    pub fn schema_update(addr:Ref, index:usize, type_id:usize);
    pub fn schema_get(addr:Ref, index:usize) -> usize;
    pub fn schema_remove(addr:Ref, index:usize);
    pub fn schema_clear(addr:Ref);
    pub fn schema_map(addr:Ref, key:usize, index:usize);
    pub fn schema_unmap(addr:Ref, key:usize);
    pub fn schema_indexof(addr:Ref, key:usize) -> usize;
    pub fn schema_keyof(addr:Ref, index:usize) -> usize;
    pub fn schema_bind(addr:Ref, id:usize) -> usize;
    pub fn schema_has(id:usize) -> bool;
    //pub fn schema_reorder(addr:Ref, from:usize, to:usize);
}
