extern "C" size_t type_outer(size_t type_id, size_t key)
{
    type_id &= ~MASK_TEMPORARY;
    return DB_TYPE.outer(type_id, key);
}

extern "C" size_t type_inner(size_t type_id)
{
    type_id &= ~MASK_TEMPORARY;
    return DB_TYPE.inner(type_id);
}

extern "C" size_t type_key(size_t type_id)
{
    type_id &= ~MASK_TEMPORARY;
    return DB_TYPE.key(type_id);
}

extern "C" size_t type_innerkey(size_t type_id)
{
    type_id &= ~MASK_TEMPORARY;
    return DB_TYPE.key(DB_TYPE.inner(type_id));
}

extern "C" size_t type_size(size_t type_id)
{
    type_id &= ~MASK_TEMPORARY;
    if(DB_TYPE.has(type_id)) {
        size_t type = DB_TYPE.key(type_id);

        switch(type) {
            case Type::TYPE_NULL:        return 0;
            case Type::TYPE_VARYING:     return sizeof(Reference);
            case Type::TYPE_BOOLEAN:     return sizeof(Type::Boolean);
            case Type::TYPE_NATURAL:     return sizeof(Type::Natural);
            case Type::TYPE_INTEGER:     return sizeof(Type::Integer);
            case Type::TYPE_SIGNIFICANT: return sizeof(Type::Significant);
            case Type::TYPE_DECIMAL:     return sizeof(Type::Decimal);
            case Type::TYPE_BLOCK:       return type_innerkey(type_id);
            case Type::TYPE_SEQUENCE:    return sizeof(Type::Sequence);

            case Type::TYPE_ARRAY: {
                size_t length = DB_TYPE.inner(type_id);
                size_t inner = DB_TYPE.inner(length);
                if(inner == Type::TYPE_NULL) {
                    return sizeof(Reference);
                }
                else {
                    return static_cast<size_t>(DB_TYPE.key(length)) * type_size(inner);
                }
            }

            case Type::TYPE_LIST:
            case Type::TYPE_SPARSE:
            case Type::TYPE_RECORD:
            {
                size_t innerkey = type_innerkey(type_id);
                if(innerkey == Type::TYPE_NULL) {
                    return sizeof(Reference);
                }
                else {
                    switch(type) {
                        case Type::TYPE_LIST: return sizeof(Type::List);
                        case Type::TYPE_SPARSE: return sizeof(Type::Sparse);
                        case Type::TYPE_RECORD: {
                            auto binding = DB_SCHEMA.get(type_innerkey(type_id));
                            if(binding != nullptr) {
                                return binding->size;
                            }
                        }
                    }
                }
            }

            case Type::TYPE_SCHEMA: return sizeof(Type::Schema);
            default: return 0;
        }
    }
    return 0;
}

size_t type_alignment(size_t type_id)
{
    type_id &= ~MASK_TEMPORARY;
    if(DB_TYPE.has(type_id)) {
        size_t type = DB_TYPE.key(type_id);

        switch(type) {
            case Type::TYPE_NULL: return 0;

            case Type::TYPE_BOOLEAN:     return sizeof(Type::Boolean);
            case Type::TYPE_NATURAL:     return sizeof(Type::Natural);
            case Type::TYPE_INTEGER:     return sizeof(Type::Integer);
            case Type::TYPE_SIGNIFICANT: return sizeof(Type::Significant);
            case Type::TYPE_DECIMAL:     return sizeof(Type::Decimal);
            case Type::TYPE_BLOCK:       return sizeof(uint8_t);
            case Type::TYPE_ARRAY:       return type_alignment(type_inner(type_id));

            case Type::TYPE_VARYING:
            case Type::TYPE_SEQUENCE:
            case Type::TYPE_LIST:
            case Type::TYPE_SPARSE:
            case Type::TYPE_SCHEMA:
                return sizeof(size_t);

            case Type::TYPE_RECORD: {
                auto binding = DB_SCHEMA.get(type_innerkey(type_id));
                return (binding != nullptr)? binding->alignment : 0;
            }
        }
    }
    return 0;
}

extern "C" size_t type_hasinner(size_t type_id)
{
    type_id &= ~MASK_TEMPORARY;
    return kind_hasinner(type_key(type_id));
}

extern "C" size_t kind_hasinner(size_t kind)
{
    switch(kind) {
        case Type::TYPE_ARRAY:
            return 2;
        case Type::TYPE_DECIMAL:
        case Type::TYPE_BLOCK:
        case Type::TYPE_LIST:
        case Type::TYPE_SPARSE:
        case Type::TYPE_RECORD:
            return 1;
    }
    return 0;
}

extern "C" size_t type_sanitize(size_t type_id)
{
    return type_id & ~MASK_TEMPORARY;
}
