#include <cstdint>
#include <vector>

extern "C" struct Reference {
    size_t   type;
    uint8_t* address;
};

namespace Type {

enum {
    TYPE_NULL = 0x00,
    TYPE_BOOLEAN = 0x01,
    /* ... */
    TYPE_BYTE = 0x04,
    TYPE_CHAR = 0x05,
    /* ... */
    TYPE_NATURAL = 0x08,
    TYPE_INTEGER = 0x09,
    TYPE_DECIMAL = 0x0A,
    TYPE_SIGNIFICANT = 0x0B,
    TYPE_COMPLEX = 0x0C,
    /* ... */
    TYPE_RANGE = 0x0F,
    TYPE_BLOCK = 0x10,
    TYPE_SEQUENCE = 0x11,
    /* ... */
    TYPE_VARYING = 0x20,
    TYPE_OPTIONAL = 0x21,
    TYPE_ARRAY = 0x22,
    TYPE_LIST = 0x23,
    TYPE_BAG = 0x24,
    TYPE_SPARSE = 0x25,
    TYPE_TRIE = 0x26,
    TYPE_MAP = 0x27,
    TYPE_TREE = 0x28,
    TYPE_GRAPH = 0x29,
    /* ... */
    TYPE_ENUM = 0x7d,
    TYPE_RECORD = 0x7e,
    TYPE_SCHEMA = 0x7f,
};

typedef bool     Boolean;
typedef uint64_t Natural;
typedef int64_t  Integer;
typedef int64_t  Decimal;
typedef double   Significant;

struct Sequence {
    RawList data;
};

struct List {
    RawList data;
};

struct PaletteList {
    struct Record {
        size_t palette;
        size_t count;
    };

    RawList data;
    RawList contents;
};

struct Table {
    RawList data;
};

extern "C" struct SparseHeader {
    size_t start;
    size_t length;
    size_t index;
};

struct Sparse {
    RawList header;
    RawList data;
};

struct Schema {
    struct Mapping {
        size_t key;
        size_t index;
        Reference object;
    };

    RawList data;
    RawList map;
};

struct SchemaBinding {
    struct Row {
        size_t key;
        size_t type;
        size_t offset;
    };

    size_t binding;
    size_t alignment;
    size_t size;
    size_t references;
    
    SparseList<size_t> map;
    std::vector<Row> data;
};

}
