macro_rules! impl_ops_numeric {
    ($clas:ty, $param:ty, $repr:ty) => {
        impl std::ops::Add<$param> for $clas {
            type Output = Self;
            fn add(self, other:$param) -> Self { Self::new().with(self.get::<$repr>() + other as $repr) }
        }
        impl std::ops::Sub<$param> for $clas {
            type Output = Self;
            fn sub(self, other:$param) -> Self { Self::new().with(self.get::<$repr>() - other as $repr) }
        }
        impl std::ops::Mul<$param> for $clas {
            type Output = Self;
            fn mul(self, other:$param) -> Self { Self::new().with(self.get::<$repr>() * other as $repr) }
        }
        impl std::ops::Div<$param> for $clas {
            type Output = Self;
            fn div(self, other:$param) -> Self { Self::new().with(self.get::<$repr>() / other as $repr) }
        }
        impl std::ops::Rem<$param> for $clas {
            type Output = Self;
            fn rem(self, other:$param) -> Self { Self::new().with(self.get::<$repr>() % other as $repr) }
        }

        impl std::ops::AddAssign<$param> for $clas {
            fn add_assign(&mut self, other:$param) { self.set(self.get::<$repr>() + other as $repr) }
        }
        impl std::ops::SubAssign<$param> for $clas {
            fn sub_assign(&mut self, other:$param) { self.set(self.get::<$repr>() - other as $repr) }
        }
        impl std::ops::MulAssign<$param> for $clas {
            fn mul_assign(&mut self, other:$param) { self.set(self.get::<$repr>() * other as $repr) }
        }
        impl std::ops::DivAssign<$param> for $clas {
            fn div_assign(&mut self, other:$param) { self.set(self.get::<$repr>() / other as $repr) }
        }
        impl std::ops::RemAssign<$param> for $clas {
            fn rem_assign(&mut self, other:$param) { self.set(self.get::<$repr>() % other as $repr) }
        }

        impl std::cmp::PartialEq<$param> for $clas {
            fn eq(&self, other:&$param) -> bool { self.get::<$repr>() == *other as $repr }
            fn ne(&self, other:&$param) -> bool { self.get::<$repr>() != *other as $repr }
        }

        impl std::cmp::PartialOrd<$param> for $clas {
            fn partial_cmp(&self, other:&$param) -> Option<std::cmp::Ordering>
            { self.get::<$repr>().partial_cmp(&(*other as $repr)) }

            fn lt(&self, other:&$param) -> bool { self.get::<$repr>() < *other as $repr }
            fn le(&self, other:&$param) -> bool { self.get::<$repr>() <= *other as $repr }
            fn gt(&self, other:&$param) -> bool { self.get::<$repr>() > *other as $repr }
            fn ge(&self, other:&$param) -> bool { self.get::<$repr>() >= *other as $repr }
        }

        impl From<$param> for $clas { fn from(value: $param) -> Self { Self::new().with(value) } }
        impl Into<$param> for $clas { fn into(self) -> $param { self.get::<$param>() } }
    }
}

macro_rules! impl_ops_numeric_self {
    ($clas:ty, $repr:ty) => {
        impl std::ops::Add<$clas> for $clas {
            type Output = Self;
            fn add(self, other:$clas) -> Self { Self::new().with(self.get::<$repr>() + other.get::<$repr>()) }
        }
        impl std::ops::Sub<$clas> for $clas {
            type Output = Self;
            fn sub(self, other:$clas) -> Self { Self::new().with(self.get::<$repr>() - other.get::<$repr>()) }
        }
        impl std::ops::Mul<$clas> for $clas {
            type Output = Self;
            fn mul(self, other:$clas) -> Self { Self::new().with(self.get::<$repr>() * other.get::<$repr>()) }
        }
        impl std::ops::Div<$clas> for $clas {
            type Output = Self;
            fn div(self, other:$clas) -> Self { Self::new().with(self.get::<$repr>() / other.get::<$repr>()) }
        }
        impl std::ops::Rem<$clas> for $clas {
            type Output = Self;
            fn rem(self, other:$clas) -> Self { Self::new().with(self.get::<$repr>() % other.get::<$repr>()) }
        }

        impl std::ops::AddAssign<$clas> for $clas {
            fn add_assign(&mut self, other:$clas) { self.set(self.get::<$repr>() + other.get::<$repr>()) }
        }
        impl std::ops::SubAssign<$clas> for $clas {
            fn sub_assign(&mut self, other:$clas) { self.set(self.get::<$repr>() - other.get::<$repr>()) }
        }
        impl std::ops::MulAssign<$clas> for $clas {
            fn mul_assign(&mut self, other:$clas) { self.set(self.get::<$repr>() * other.get::<$repr>()) }
        }
        impl std::ops::DivAssign<$clas> for $clas {
            fn div_assign(&mut self, other:$clas) { self.set(self.get::<$repr>() / other.get::<$repr>()) }
        }
        impl std::ops::RemAssign<$clas> for $clas {
            fn rem_assign(&mut self, other:$clas) { self.set(self.get::<$repr>() % other.get::<$repr>()) }
        }

        impl std::cmp::PartialEq<$clas> for $clas {
            fn eq(&self, other:&$clas) -> bool { self.get::<$repr>() == other.get::<$repr>() }
            fn ne(&self, other:&$clas) -> bool { self.get::<$repr>() != other.get::<$repr>() }
        }

        impl std::cmp::PartialOrd<$clas> for $clas {
            fn partial_cmp(&self, _other:&$clas) -> Option<std::cmp::Ordering>
            { None }

            fn lt(&self, other:&$clas) -> bool { self.get::<$repr>() < other.get::<$repr>() }
            fn le(&self, other:&$clas) -> bool { self.get::<$repr>() <= other.get::<$repr>() }
            fn gt(&self, other:&$clas) -> bool { self.get::<$repr>() > other.get::<$repr>() }
            fn ge(&self, other:&$clas) -> bool { self.get::<$repr>() >= other.get::<$repr>() }
        }
    }
}

macro_rules! impl_ops_numeric_all {
    ($clas:ty, $repr:ty) => {
        use crate::util::macros::{impl_ops_numeric, impl_ops_numeric_self};
        impl_ops_numeric!($clas, i128, $repr);
        impl_ops_numeric!($clas, i64, $repr);
        impl_ops_numeric!($clas, i32, $repr);
        impl_ops_numeric!($clas, i16, $repr);
        impl_ops_numeric!($clas, i8, $repr);
        impl_ops_numeric!($clas, u128, $repr);
        impl_ops_numeric!($clas, u64, $repr);
        impl_ops_numeric!($clas, u32, $repr);
        impl_ops_numeric!($clas, u16, $repr);
        impl_ops_numeric!($clas, u8, $repr);
        impl_ops_numeric!($clas, f64, $repr);
        impl_ops_numeric!($clas, f32, $repr);
        impl_ops_numeric_self!($clas, $repr);
    }
}

pub(crate) use impl_ops_numeric;
pub(crate) use impl_ops_numeric_self;
pub(crate) use impl_ops_numeric_all;
